local Raycaster = {}

function Raycaster:init(map, num_rays, ...)

    local screen_width = love.graphics.getWidth()
    local screen_height = love.graphics.getHeight()

    local h_screen_width = screen_width * .5
    local h_screen_height = screen_height * .5

    num_rays = num_rays or screen_width -- Pour une scene 3D nous avons besoin du même nombre de rayon que de pixels de largeur

    -- Create or reset result table --

    if not self.result then
        self.result = {}
    else
        local r = self.result
        r.px, r.py = nil, nil
        for i = 1, #r do r[i] = nil end
    end

    -- Init new values --

    self:setMap(map)
    self.num_rays = num_rays
    self.screen_width = screen_width
    self.screen_height = screen_height
    self.h_screen_width = h_screen_width
    self.h_screen_height = h_screen_height

end

function Raycaster:setMap(map)
    self.map_width = #map[1]
    self.map_height = #map
    self.map = map
end

function Raycaster:posToMap(x,y)
    return math.floor(x), math.floor(y)
end

function Raycaster:setTextures(...)

    local textures = {}
    local n = select("#", ...)

    for i = 1, n do
        local v = select(i, ...)
        if type(v) == "string" then
            v = love.image.newImageData(v)
        end
        textures[i] = v
    end

    self.textures = textures

end

function Raycaster:getPlaneFromDir(dir_x, dir_y)    -- Prend en entrée une direction pour renvoyé son vecteur sur le plan orthogonal de -0.66 à 0.66 (getPerpendicularVector)
    local length = math.sqrt(dir_x*dir_x + dir_y*dir_y)
    return (dir_y/length)*.66, (dir_x/length)*.66
end

function Raycaster:update(px, py, dir_x, dir_y, plane_x, plane_y)

    -- Define local self values --

    local num_rays = self.num_rays
    local screen_height = self.screen_height
    local h_screen_height = self.h_screen_height

    local map = self.map

    local r = self.result
    r.px, r.py = px, py

    -- Perform raycasting --

    local j = 1 -- i: 11 by 11

    for i = 1, num_rays do

        local map_x, map_y = self:posToMap(px,py)   -- Position initial du rayon dans le tableau map

        local cam_x = 2 * i / num_rays - 1          -- Coordonnée x dans l'espace de la caméra
        local ray_dir_x = dir_x + plane_x * cam_x
        local ray_dir_y = dir_y + plane_y * cam_x

        -- On obtiens la distance jusqu'au bord de la case suivante --

        local dx = (ray_dir_x == 0) and 1e30 or math.abs(1/ray_dir_x)
        local dy = (ray_dir_y == 0) and 1e30 or math.abs(1/ray_dir_y)

        -- On calcul l'etape initial et la longueur jusqu'au côté --

        local step_x, side_dist_x;
        local step_y, side_dist_y;

        if ray_dir_x < 0 then step_x, side_dist_x = -1, (px-map_x)*dx
        else step_x, side_dist_x = 1, (map_x+1-px)*dx end

        if ray_dir_y < 0 then step_y, side_dist_y = -1, (py-map_y)*dy
        else step_y, side_dist_y = 1, (map_y+1-py)*dy end

        -- On lance le rayon --

        local side; repeat  -- DDA

            if side_dist_x < side_dist_y then
                side_dist_x = side_dist_x + dx
                map_x = map_x + step_x
                side = 0
            else
                side_dist_y = side_dist_y + dy
                map_y = map_y + step_y
                side = 1
            end

        until not map[map_y] or map[map_y][map_x] ~= 0;

        -- Ajout des données du rayon dans la table result --

        local distance_to_wall = (side == 0) and side_dist_x - dx or side_dist_y - dy
        local wall_height = screen_height / distance_to_wall -- Calculation of the wall height
        local wall_height_half = wall_height * .5

        r[j]    = px + ray_dir_x * distance_to_wall             -- HIT X
        r[j+1]  = py + ray_dir_y * distance_to_wall             -- HIT Y
        r[j+2]  = ray_dir_x                                     -- RAY DIR X
        r[j+3]  = ray_dir_y                                     -- RAY DIR Y
        r[j+4]  = distance_to_wall                              -- DIST TO WALL
        r[j+5]  = i                                             -- WALL X
        r[j+6]  = -wall_height_half + h_screen_height           -- WALL Y1
        r[j+7]  = wall_height_half + h_screen_height            -- WALL Y2
        r[j+8]  = wall_height                                   -- WALL H
        r[j+9]  = map[map_y][map_x]                             -- texNum
        r[j+10] = side                                          -- side

        -- We go to the next angle step (next ray) --

        j = j + 11

    end

end

function Raycaster:renderMap(x,y,w,h,tile_size)

    tile_size = tile_size or 32

    -- Define local self values --

    local map = self.map
    local map_width = self.map_width
    local map_height = self.map_height

    map_width = map_width * tile_size
    map_height = map_height * tile_size

    -- Define local result values --

    local r = self.result
    local px, py = r.px-1, r.py-1

    love.graphics.push()
    love.graphics.translate(x,y)
    love.graphics.scale(w/map_width, h/map_height)

        love.graphics.setColor(.5,.5,.5)
        love.graphics.rectangle("fill", 0, 0, map_width, map_height)

        -- Draw map --

        love.graphics.setColor(.125,.125,.125)

        for ty, xs in ipairs(map) do
            for tx, value in ipairs(xs) do
                if value ~= 0 then
                    love.graphics.rectangle("fill", (tx-1)*tile_size, (ty-1)*tile_size, tile_size, tile_size)
                end
            end
        end

        -- Draw rays --

        love.graphics.setColor(1,0,0)

        for i = 1, #r-10, 11 do
            love.graphics.line(px*tile_size, py*tile_size, (r[i]-1)*tile_size, (r[i+1]-1)*tile_size)
        end

        -- Draw player position --

        love.graphics.setColor(1,1,0)
        love.graphics.circle("fill", px*tile_size, py*tile_size, tile_size*.25)

    love.graphics.pop()

end

function Raycaster:renderView()

    local r = self.result
    local textures = self.textures
    local screen_width = self.screen_width
    local screen_height = self.screen_height
    local h_screen_height = self.h_screen_height

    love.graphics.setColor(0,.1,.5)
    love.graphics.rectangle("fill", 0,0,screen_width,h_screen_height)

    love.graphics.setColor(0,.5,.1)
    love.graphics.rectangle("fill", 0,h_screen_height,screen_width,h_screen_height)

    for i = 1, #r-10, 11 do

        local x = r[i+5]
        local y1 = r[i+6]
        local y2 = r[i+7]
        local h = r[i+8]

        local texNum = r[i+9]
        local side = r[i+10]

        local c = h/200
        local tex = textures and textures[texNum]

        if tex then

            local ray_dir_x = r[i+2]
            local ray_dir_y = r[i+3]

            local tex_width = tex:getWidth()
            local tex_height = tex:getHeight()

            -- calculate value of wall_x
            local wall_x -- where exactly the wall was hit
            if side == 0 then
                wall_x = r[i+1]
            else
                wall_x = r[i]
            end
            wall_x = wall_x - math.floor(wall_x)

            -- x coordinate on the texture
            local tex_x = math.floor(wall_x * tex_width)
            if side == 0 and ray_dir_x > 0 then
                tex_x = tex_width - tex_x - 1
            end
            if side == 1 and ray_dir_y < 0 then
                tex_x = tex_width - tex_x - 1
            end

            -- How much to increase the texture coordinate per screen pixel
            local step = 1.0 * tex_height / h

            -- Starting texture coordinate
            local tex_pos = (y1 - screen_height / 2 + h / 2) * step

            for y = y1, y2 do

                local tex_y = math.floor(tex_pos) % tex_height
                tex_pos = tex_pos + step

                local cr,cg,cb = tex:getPixel(tex_x,tex_y)
                love.graphics.setColor(cr*c,cg*c,cb*c)
                love.graphics.points(x,y)

            end

        else
            love.graphics.setColor(c,c,c)
            love.graphics.setLineStyle("rough")
            love.graphics.line(x, y1, x, y2)
            love.graphics.setLineStyle("smooth")
        end

    end

end

function Raycaster:resize(w,h,adjust_num_rays)

    if adjust_num_rays then

        local old_num_rays = self.num_rays
        self.num_rays = w * (self.num_rays/self.screen_width)

        if old_num_rays > self.num_rays then

            local r = self.result

            for i = self.num_rays, old_num_rays do

                local j = i*11

                for k = 0, 10 do
                    r[j+k] = nil
                end

            end

        end

    end

    self.screen_width, self.screen_height = w, h
    self.h_screen_width, self.h_screen_height = w*.5, h*.5

end


return Raycaster